//
//  main_chirp_v3.0.c
//
//  Hideyuki Tagoshi 2017/07/15
//  
//  This code is made for the mini-school at NTNU, Taiwan (2017/7/13-15)
//
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <complex.h>
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <fftw3.h>

#define PI            3.14159265358979323846
#define SOLARTIME     (4.925491e-6)   // [sec]
#define TTSOLAR       (4.925491e-6)   // [sec]
#define CLIGHT        (2.99792458e10) // [cm/s]
#define Mpc           (3.08568025e24) // [cm]
#define REARTH        (6378.137e5)    // [cm]
#define DATASCALE     (1.e20)  // 

// (tau0,tau3) => (Mc,eta)
void tau03toMceta(double tau0, double tau3, double fa, double *Mc, double *eta)
{
    double tmp1,tmp2;
    
    tmp1 = pow(PI*fa,-8./5);
    tmp2 = pow(PI*fa,-25./9);
    *Mc  = pow(5./256/tau0,3./5)*tmp1/TTSOLAR;
    *eta = pow(PI/8./tau3,5./3)*pow(*Mc*TTSOLAR,-10./9)*tmp2;
    
    return;
}

// (tau0,tau3) => (Mt,eta)
void tau03toMteta(double tau0, double tau3, double fa, double *Mt, double *eta)
{
    
    *Mt  = 5./32/PI*tau3/tau0/PI/fa/TTSOLAR;
    *eta = pow(5.,-2./3)*pow(2.,1./3)*pow(PI*tau0/tau3,2./3)/tau3/fa;
    
    return;
}

// 1PN post-Newtonian phase with phic in frequency domain
double phasephi(double f, double Mt, double eta, double phic)
{
    double x;
    
    x = pow(PI*TTSOLAR*Mt*f,1./3.);
    
    return -3./128./eta/pow(x,5.)*(1+(3715./756.+55./9*eta)*pow(x,2.))+phic;
    
}

// 1PN post-Newtonian phase with tc and phic in frequency domain
double phasetcphi(double f, double Mt, double eta, double tc, double phic)
{
    double x;
    
    x = pow(PI*TTSOLAR*Mt*f,1./3.);
    
    return -3./128./eta/pow(x,5.)*(1+(3715./756.+55./9*eta)*pow(x,2.))+phic-2*PI*f*tc;
    
}


/* freqpow73[0..npoint/2]= f^(-7/3) */
void norm_template(int npoint,  double *Sn,  double *freqpow73,
                   double fmin, double fmax, double Mt, double *norm)
{
    int i,n,imin,imax;
    double nnorm,df,fisco,maxf;
    
    n    = npoint/2;
    df   = fmax/n;
    imin = fmin/df;
    
    fisco = 1./(pow(6,1.5)*PI*Mt*TTSOLAR);
    
    if(fmax>fisco) maxf = fisco;
    else           maxf = fmax;
    
    imax = maxf/df;
    
    nnorm = 0.5*(freqpow73[imin]/Sn[imin]+freqpow73[imax]/Sn[imax]);
    
    for( i = imin+1; i < imax; i++)
    {
        nnorm += freqpow73[i]/Sn[i];
    }
    nnorm = nnorm*df;
    *norm = pow(4*nnorm,0.5);
    
}

// Theoretical noise power spectrum density of TAMA phase II.
// Overall factor 2.4/32*1.e-43 is omitted.
void noisePSD(int n, double *Sn, double fmin, double fmax)
// n=npoint/2, fmax=srate/2
{
    int i;
    double freq,df;
    
    df=fmax/n;
    
    Sn[0]=0;
    for(i = 1; i <= n; i++){
        freq  = df*i;
        Sn[i] = (2.4/32.*1.e-43)*(pow(95./freq,25.) + 13.*pow(220./freq,4.) 
                                  + 9*(1 + pow(freq/250,2.)));
    }
    
}


void generate_gaussnoise_gsl(int npoint, double *Sn, double fmin, double fmax,
                             double *noise, gsl_rng *r)
{
    double gsl_ran_gaussian(const gsl_rng * rr,const double sigma);
    int k,kmin,n;
    double df,srate,sig;
    
    n = npoint/2;
    
    df    = fmax/n;
    kmin  = (int)(fmin/df);
    srate = fmax*2;
    
    for(k = 0; k < kmin; k++)
    {
        noise[2*k]   = 0;
        noise[2*k+1] = 0;
    }
    
    for(k = kmin; k < npoint/2; k++)
    {
        sig = pow(npoint/srate*Sn[k]/2/2,0.5);
        noise[2*k]   = gsl_ran_gaussian(r,sig);
        noise[2*k+1] = gsl_ran_gaussian(r,sig);
    }
    
    sig = pow(npoint/srate*Sn[npoint/2]/2/2,0.5);
    noise[1] = gsl_ran_gaussian(r,sig);
    
    return;
}

/*----------------------------------------------------------------------------*/

int main(int argc,char *argv[])
{
    int i,k,npoint,kmin;
    double freq,fmin,fmax,srate,df,tmp;
    const gsl_rng_type * T;
    gsl_rng * r;
    
    npoint    = 65536;          // number of data point
    fmin      = 60;             // lowest frequency [Hz]
    srate     = 4096;           // Sampling rate [Hz]. The length of data is npoint/srate [sec]
    fmax      = srate/2;        // upper cutoff frequency (Nyquist frequency) [Hz]
    df        = srate/npoint;   // frequency resolution [Hz]
    kmin      = (int)(fmin/df); 
    
    // Signal parameters
    double SNR =  10; // SNR of signal
    double m1s =  1.4;
    double m2s =  1.4;
    double tcs =  12;
    double phics= 0;

    // Template parameters
    double m1t   = 1.4; 
    double m2t   = 1.4; 
    
    ////////////////////////////////////////////////////////////
    // Setting up for the GSL random number generator
    // The seed is changed everytime for this example. 
    // If you want to give the seed manually, give "seed" some fixed number.
    ////////////////////////////////////////////////////////////
    unsigned long int seed=(unsigned)time(NULL);
    gsl_rng_env_setup();
    T = gsl_rng_default;
    r = gsl_rng_alloc (T);
    gsl_rng_set(r,seed);
    ////////////////////////////////////////////////////////////
    // If you want to use the default value of seed, use below.
    ////////////////////////////////////////////////////////////
    //  gsl_rng_env_setup();
    //  T = gsl_rng_default;
    //  r = gsl_rng_alloc (T);
    ////////////////////////////////////////////////////////////
    
    double *Sn            = NULL; 
    double *noise_temp    = NULL;
    double complex *noise = NULL;
    
    Sn         = (double *)malloc((size_t)((npoint/2+1)*sizeof(double)));
    noise_temp = (double *)malloc((size_t)((npoint)*sizeof(double)));
    noise      = (double complex *)malloc((size_t)((npoint/2+1)*sizeof(double complex)));
    
    // Define analytical one-sided noise power spectrum density
    noisePSD(npoint/2,Sn,fmin,fmax);
    
    // Generate Gaussian noise in frequency domain
    generate_gaussnoise_gsl(npoint,Sn,fmin,fmax,noise_temp,r);
    
    // Store Gaussian noise in the complex double variable, noise[]
    noise[0]        = noise_temp[0];
    noise[npoint/2] = noise_temp[1];
    
    for(k = 1; k < npoint/2; k++)
    {
        noise[k] = noise_temp[2*k] + noise_temp[2*k+1]*I;
    }    
    
    FILE *fp;
    fp=fopen("spectrum.dat","w");
    for(k=kmin; k<=npoint/2; k++)
    {
        tmp = cabs(noise[k])*cabs(noise[k]);
        tmp = tmp/(npoint/srate)*2;
        fprintf(fp,"%e %e %e %e %e\n",df*k,creal(noise[k]),cimag(noise[k]),
                sqrt(tmp),sqrt(Sn[k]));
    }
    fclose(fp);
    
    
    /*=============================================================*/
    /* Signal generation */
    double Mts,etas;
    
    Mts  = m1s + m2s;
    etas = m1s*m2s/Mts/Mts;
    
    double maxf;
    double fisco = 1./(pow(6,1.5)*PI*Mts*TTSOLAR);
    
    if(fmax>fisco) maxf = fisco;
    else           maxf = fmax;
    
    int maxk = (int)(maxf/df);
    
    double complex *data     = NULL;
    double complex *signal   = NULL;
    double *freqpow73        = NULL;
    double *freqpow76        = NULL;
    
    data      = (double complex *)malloc((size_t)((npoint/2+1)*sizeof(double complex)));
    signal    = (double complex *)malloc((size_t)((npoint/2+1)*sizeof(double complex)));
    freqpow73 = (double *)malloc((size_t)((npoint/2+1)*sizeof(double)));
    freqpow76 = (double *)malloc((size_t)((npoint/2+1)*sizeof(double)));
    
    freqpow73[0] = 0;
    freqpow76[0] = 0;
    
    for(k = 1; k <= npoint/2; k++)
    {
        freq         = k*df;
        freqpow73[k] = pow(freq,-7./3);
        freqpow76[k] = pow(freq,-7./6);
    }
    
    double norms;
    norm_template(npoint,Sn,freqpow73,fmin,fmax,Mts,&norms);  
    
    for(i = 0; i <= npoint/2; i++)
    {
        data[i]     = 0;
        signal[i]   = 0;
    }  
    
    double tempR, tempI;
    for(k = kmin; k < maxk; k++)
    {
        freq      = k*df;
        tempR     = SNR/norms*freqpow76[k]*cos(phasetcphi(freq,Mts,etas,tcs,phics));
        tempI     = SNR/norms*freqpow76[k]*sin(phasetcphi(freq,Mts,etas,tcs,phics));
        signal[k] = tempR + tempI * I;
    }
    
    for(k = 0; k <= npoint/2; k++)
    {
        data[k] = noise[k] + signal[k];
    }
/*    
    fp = fopen("spectrum_chirp.dat","w");
    
    for(k = 1; k < npoint/2; k++)
        fprintf(fp,"%e %e %e\n",k*df, signal[2*k], signal[2*k+1]);
    
    fclose(fp);
*/    

/*=====================================================*/
/* Generate noise+signal data in time domain           */
    
/* Prepare input and out arrays for FFTW3 */
/* We use the DFT of "complex to real" value */
    fftw_plan p,p1;
    fftw_complex *inc2r,*inc2r_2;
    
    double *outc2r = NULL;
    double *timeseries_data   = NULL;
    double *timeseries_signal = NULL;
    
    size_t mem_size   = sizeof(fftw_complex) * (npoint/2+1);
    inc2r             = (fftw_complex*)fftw_malloc( mem_size );
    inc2r_2           = (fftw_complex*)fftw_malloc( mem_size );
    outc2r            = (double *)malloc((size_t)(npoint*sizeof(double)));
    timeseries_data   = (double *)malloc((size_t)(npoint*sizeof(double)));
    timeseries_signal = (double *)malloc((size_t)(npoint*sizeof(double)));

    /*---------------------------------------------------------------*/

    p  = fftw_plan_dft_c2r_1d( npoint, inc2r, timeseries_data, FFTW_ESTIMATE );

    /* Give the data to input array, and execute FFTW3 */
    /* The data must be substituted after fftw_plan_ */
    for ( k = 0; k <= npoint/2; k++ )
        inc2r[k] = data[k];
        
    fftw_execute(p);

    /*---------------------------------------------------------------*/

    p1 = fftw_plan_dft_c2r_1d( npoint, inc2r_2, timeseries_signal, FFTW_ESTIMATE );

    for ( k = 0; k <= npoint/2; k++ )
        inc2r_2[k] = signal[k];
    
    fftw_execute(p1);

    /*---------------------------------------------------------------*/
        
    fp=fopen("timeseries_noise_and_chirp.dat","w");
    for(i=0;i<npoint;i++)
    {
        timeseries_data[i]   = timeseries_data[i]*df;
        timeseries_signal[i] = timeseries_signal[i]*df;
        fprintf(fp,"%e %e %e\n",i/srate,timeseries_data[i],timeseries_signal[i]);
    }
    fclose(fp);
    
    fftw_destroy_plan(p);  
    fftw_destroy_plan(p1);
    fftw_free(inc2r);
    fftw_free(inc2r_2);
    free(outc2r);
    free(timeseries_signal);
    
    /*=============================================================*/
    /*=============================================================*/

    /* Matched filtering */
    /* Prepare template */
    double Mtt,etat,phict,normt;
    
    /* mass and phase are the same as the signal.
     Results are given for different tc */
    phict = 0.;
    Mtt   = m1t+m2t; 
    etat  = m1t*m2t/Mtt/Mtt; 
    
    norm_template(npoint,Sn,freqpow73,fmin,fmax,Mtt,&normt);
    
//    printf("%e\n",normt);
    
    double complex *templateC = NULL;
    double complex *templateS = NULL;
    
    templateC  = (double complex *)malloc((size_t)((npoint/2+1)*sizeof(double complex)));
    templateS  = (double complex *)malloc((size_t)((npoint/2+1)*sizeof(double complex)));
    
    for(i = 0; i <= npoint/2; i++)
    {
        templateC[i] = 0;
        templateS[i] = 0;
    }  
    
    for( k=1; k <= npoint/2; k++ )
    {
        freq         = k*df;
        tempR        = 1./normt*freqpow76[k]*cos(phasephi(freq,Mtt,etat,phict));
        tempI        = 1./normt*freqpow76[k]*sin(phasephi(freq,Mtt,etat,phict));
        templateC[k] = tempR + tempI * I;
        templateS[k] = templateC[k]*I;
    }
    
    /* Prepare input and out arrays for FFTW3 */
    /* We use the DFT of "complex to real" value */

    double *outc2r_C = NULL;
    double *outc2r_S = NULL;
    
    fftw_plan p2, p3;
    size_t mem_size2 = sizeof(fftw_complex) * (npoint/2+1);
    inc2r            = (fftw_complex*)fftw_malloc( mem_size2 );
    outc2r_C           = (double *)malloc((size_t)(npoint*sizeof(double)));
    outc2r_S           = (double *)malloc((size_t)(npoint*sizeof(double)));

    p2 = fftw_plan_dft_c2r_1d( npoint, inc2r, outc2r_C, FFTW_ESTIMATE );
    
    /* Give the data to input array, and execute FFTW3 */
    /* The data must be substituted after fftw_plan_ */
    /* Integrand of matched filter */
    
    inc2r[0] = 0;
    
    for( k = 1; k <= npoint/2; k++ )
    {
        inc2r[k] = 2 * data[k] * conj(templateC[k]) / Sn[k];
    }

    fftw_execute(p2);
    
    p3 = fftw_plan_dft_c2r_1d( npoint, inc2r, outc2r_S, FFTW_ESTIMATE );
    inc2r[0] = 0;
    for( k = 1; k <= npoint/2; k++ )
    {
        inc2r[k] = 2 * data[k] * conj(templateS[k]) / Sn[k];
    }
    
    fftw_execute(p3);    
    
    double rho;
    
    fp=fopen("result_MF_chirp_tc.dat","w");
    for(k=0;k<npoint;k++)
    {
        outc2r_C[k] = outc2r_C[k]*df;
        outc2r_S[k] = outc2r_S[k]*df;
        rho=sqrt(outc2r_C[k]*outc2r_C[k]+outc2r_S[k]*outc2r_S[k]);
        fprintf(fp,"%e %e\n",k/srate,rho);
    }
    fclose(fp);
    
    
    fftw_destroy_plan(p2);
    fftw_destroy_plan(p3);
    fftw_free(inc2r);
    free(outc2r_C);
    free(outc2r_S);
    
    /*=============================================================*/

    if(Sn        != NULL) free(Sn);
    if(noise_temp!= NULL) free(noise_temp);
    if(noise     != NULL) free(noise);
    if(timeseries_data != NULL) free(timeseries_data);
    if(data      != NULL) free(data);
    if(signal    != NULL) free(signal);
    if(templateC != NULL) free(templateC);
    if(templateS != NULL) free(templateS);
    if(freqpow76 != NULL) free(freqpow76);
    free(freqpow73);
        
    return 0;
}
